/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.context.servlet;

import net.craftforge.commons.collections.CaseInsensitiveMap;
import net.craftforge.essential.context.Request;
import net.craftforge.essential.core.constants.HttpMethods;

import javax.inject.Inject;
import javax.servlet.http.HttpServletRequest;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;

/**
 * Implementation of a request for servlets.
 * Uses the {@link HttpServletRequest} as source of request
 * information.
 *
 * @author Christian Bick
 * @since 26.12.11
 */
public class ServletWebRequest implements Request {

    private HttpServletRequest request;

    private Map<String, String[]> headers;

    @Inject
    public ServletWebRequest(HttpServletRequest request) {
        this.request = request;
    }

    /**
     * {@inheritDoc}
     */
    public String getHttpMethod() {
        return request.getMethod();
    }

    /**
     * {@inheritDoc}
     */
    public String getUrlResourcePart() {
        String urlResourcePart = "";
        String servletPath = request.getServletPath();
        String pathInfo = request.getPathInfo();
        if (servletPath != null && ! servletPath.isEmpty()) {
            if (! servletPath.startsWith("/")) {
                urlResourcePart += "/";
            }
            urlResourcePart += servletPath;
        }
        if (pathInfo != null && ! pathInfo.isEmpty()) {
            if (! pathInfo.startsWith("/") && ! urlResourcePart.equals("/")) {
                urlResourcePart += "/";
            }
            urlResourcePart += pathInfo;
        }
        return urlResourcePart;
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public Map<String, String[]> getHeaders() {
        if (headers != null) {
            return headers;
        }
        headers = new CaseInsensitiveMap<String[]>();
        for (String headerName : (List<String>) Collections.list(request.getHeaderNames())) {
            Enumeration<String> headerValues = request.getHeaders(headerName);
            List<String> headerValueList = Collections.list(headerValues);
            headers.put(headerName, headerValueList.toArray(new String[headerValueList.size()]));
        }
        return headers;
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public Map<String, String[]> getParameters() {
        return request.getParameterMap();
    }

    /**
     * {@inheritDoc}
     */
    public InputStream getBodyInputStream() {
        if (getHttpMethod().equals(HttpMethods.PUT)
                || getHttpMethod().equals(HttpMethods.POST)) {
            try {
                return request.getInputStream();
            } catch (IOException e) {
                throw new RuntimeException("Unable to resolve request body input stream", e);
            }
        }
        return null;
    }
}
