/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.controller.phases;

import net.craftforge.essential.context.Application;
import net.craftforge.essential.controller.ControllerConfiguration;
import net.craftforge.essential.controller.ControllerException;
import net.craftforge.essential.controller.ControllerPhase;
import net.craftforge.essential.controller.ControllerState;
import net.craftforge.essential.controller.resolvers.InputResolver;
import net.craftforge.essential.controller.resolvers.MediaTypeResolver;
import net.craftforge.essential.controller.resolvers.MethodResolver;
import net.craftforge.essential.controller.utils.ControllerReflUtils;
import net.craftforge.essential.core.constants.HttpMethods;
import net.craftforge.essential.core.constants.HttpStatusCode;
import net.craftforge.essential.supply.Consumer;
import net.craftforge.essential.supply.designators.ConsumerDesignator;
import net.craftforge.essential.supply.exceptions.BadInputException;
import net.craftforge.essential.supply.exceptions.UnsupportedMediaTypeException;
import net.craftforge.reflection.utils.PropertyUtils;
import net.craftforge.reflection.utils.ReflUtils;

import javax.inject.Inject;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.Date;
import java.util.List;

/**
 * Creates an instance of the resource class, initializes annotated class properties
 * and invokes the resource method. Sets the result of the resource method.
 *
 * @author Christian Bick
 * @since 03.02.2011
 */
public class InvocationPhase implements ControllerPhase {

    private ControllerConfiguration configuration;
    private MethodResolver methodResolver;
    private MediaTypeResolver mediaTypeResolver;
    private InputResolver inputResolver;
    private Application application;

    @Inject
    public InvocationPhase(ControllerConfiguration configuration, MethodResolver methodResolver,
                           MediaTypeResolver mediaTypeResolver, InputResolver inputResolver,
                           Application application) {
        this.configuration = configuration;
        this.methodResolver = methodResolver;
        this.mediaTypeResolver = mediaTypeResolver;
        this.inputResolver = inputResolver;
        this.application = application;
    }

    /**
     * {@inheritDoc}
     */
    public void run(ControllerState state) throws ControllerException {
        // Try to get the consumer class from an annotation, except in case of OPTIONS
        List<Class<?>> consumerClasses = null;
        if (! methodResolver.getHttpMethod().equals(HttpMethods.OPTIONS)) {
            consumerClasses = ControllerReflUtils.getConsumersFromMethodOrClass(state.getResourceMethod());
        }

        // In case consumer classes were not set via annotation, then take the default ones
        if (consumerClasses == null) {
            consumerClasses = configuration.getDefaultConsumerClasses();
        }

        // Designate the (best matching) producer and media type
        ConsumerDesignator consumerDesignator = ConsumerDesignator.getInstance(consumerClasses);
        Consumer consumer;
        try {
            String mediaType = mediaTypeResolver.getContentMediaType();
            consumer = consumerDesignator.designate(new String[] { mediaType });
        } catch (UnsupportedMediaTypeException e) {
            throw new ControllerException("The given content type is not supported", HttpStatusCode.UnsupportedMediaType);
        }

        // Obtain resource object
        Object resource = application.provideInstance(state.getResourceClass());

        for (Field field : ControllerReflUtils.getAnnotatedFieldsFromClass(resource.getClass())) {
            Object input;
            try {
                input = inputResolver.getInputForField(field, consumer);
            } catch (BadInputException e) {
                throw new ControllerException("The given content is not valid", e, HttpStatusCode.BadRequest);
            } catch (UnsupportedMediaTypeException e) {
                throw new ControllerException("The given content media type is not supported", e, HttpStatusCode.UnsupportedMediaType);
            } catch (UnsupportedEncodingException e) {
                throw new ControllerException("The given content charset encoding is not supported", e, HttpStatusCode.UnsupportedMediaType);
            }

            try {
                PropertyUtils.setProperty(resource, field, input);
            } catch (NoSuchMethodException e) {
                throw new ControllerException("A property of the resource class could not be" +
                        " set via the setter method: " + ReflUtils.setterName(field.getName()) + "()" +
                        " in resource class " + state.getResourceClass().getName(), e);
            } catch (InvocationTargetException e) {
                throw new ControllerException("A property of the resource class could not be" +
                        " set via the setter method: " + ReflUtils.setterName(field.getName()) + "()" +
                        " in resource class " + state.getResourceClass(), e.getCause());
            } catch (IllegalAccessException e) {
                throw new ControllerException("A property of the resource class could not be" +
                        " set: " + field.getName() +
                        " in resource class " + state.getResourceClass(), e);
            }
        }

        // Obtain method parameter values
        List<Object> parameterValues;
        try {
            parameterValues = inputResolver.getInputForMethodParameters(state.getResourceMethod(), consumer);
        } catch (BadInputException e) {
            throw new ControllerException("The given content is not valid", e, HttpStatusCode.BadRequest);
        } catch (UnsupportedMediaTypeException e) {
            throw new ControllerException("The given content media type is not supported", e, HttpStatusCode.UnsupportedMediaType);
        } catch (UnsupportedEncodingException e) {
            throw new ControllerException("The given content charset encoding is not supported", e, HttpStatusCode.UnsupportedMediaType);
        }

        // Invoke resource method
        Object result = null;
        state.setInvocationStart(new Date());
        try {
            if (! state.getResourceMethod().getReturnType().equals(void.class)) {
                if (parameterValues.size() == 0) {
                    result = state.getResourceMethod().invoke(resource);
                } else {
                    result = state.getResourceMethod().invoke(resource, parameterValues.toArray());
                }
                state.setStatus(HttpStatusCode.OK);
            } else {
                if (parameterValues.size() == 0) {
                    state.getResourceMethod().invoke(resource);
                } else {
                    state.getResourceMethod().invoke(resource, parameterValues.toArray());
                }
                state.setStatus(HttpStatusCode.NoContent);
            }
        } catch (InvocationTargetException e) {
            Throwable cause = e.getCause();
            if (cause instanceof ControllerException) {
                throw (ControllerException)cause;
            }
            throw new ControllerException("The resource method failed during execution: " +
                    "" + state.getResourceMethod(), cause);
        } catch (IllegalAccessException e) {
            throw new ControllerException("The resource method could not be invoked: " +
                    "" + state.getResourceMethod(), e);
        }
        state.setInvocationEnd(new Date());
        state.setResult(result);
    }
}
