/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.controller.phases;

import net.craftforge.essential.context.Response;
import net.craftforge.essential.controller.ControllerConfiguration;
import net.craftforge.essential.controller.ControllerException;
import net.craftforge.essential.controller.ControllerPhase;
import net.craftforge.essential.controller.ControllerState;
import net.craftforge.essential.controller.resolvers.CharsetResolver;
import net.craftforge.essential.controller.resolvers.MediaTypeResolver;
import net.craftforge.essential.controller.resolvers.MethodResolver;
import net.craftforge.essential.controller.utils.ControllerReflUtils;
import net.craftforge.essential.core.constants.HttpHeaders;
import net.craftforge.essential.core.constants.HttpMethods;
import net.craftforge.essential.core.constants.HttpStatusCode;
import net.craftforge.essential.supply.Producer;
import net.craftforge.essential.supply.designators.ProducerDesignator;
import net.craftforge.essential.supply.exceptions.UnsupportedMediaTypeException;

import javax.inject.Inject;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Produces a serialized response from the result.
 *
 * @author Christian Bick
 * @since 17.01.2011
 */
public class ProductionPhase implements ControllerPhase {

    private Response response;
    private ControllerConfiguration configuration;
    private MethodResolver methodResolver;
    private MediaTypeResolver mediaTypeResolver;
    private CharsetResolver charsetResolver;

    @Inject
    public ProductionPhase(Response response, ControllerConfiguration configuration, MethodResolver methodResolver,
                           MediaTypeResolver mediaTypeResolver, CharsetResolver charsetResolver) {
        this.response = response;
        this.configuration = configuration;
        this.methodResolver = methodResolver;
        this.mediaTypeResolver = mediaTypeResolver;
        this.charsetResolver = charsetResolver;
    }

    /**
     * {@inheritDoc}
     */
    public void run(ControllerState state) throws ControllerException {
        if (state.getResult() == null) {
            // If the result is null, only send headers
            response.sendHeaders(state.getStatus(), getHeadersWithoutContent());
        } else {
            // Try to get the producer class from an annotation, except in case of OPTIONS
            List<Class<?>> producerClasses = null;
            if (! methodResolver.getHttpMethod().equals(HttpMethods.OPTIONS)) {
                producerClasses = ControllerReflUtils.getProducersFromMethodOrClass(state.getResourceMethod());
            }

            // In case producer classes were not set via annotation, then take the default ones
            if (producerClasses == null) {
                producerClasses = configuration.getDefaultProducerClasses();
            }

            // Designate the (best matching) producer and media type
            ProducerDesignator producerDesignator = ProducerDesignator.getInstance(producerClasses);
            Producer producer;
            String mediaType;
            try {
                String[] acceptedMediaTypes = mediaTypeResolver.getAcceptedMediaTypes();
                producer = producerDesignator.designate(acceptedMediaTypes);
                mediaType = producerDesignator.designateMediaType(acceptedMediaTypes);
            } catch (UnsupportedMediaTypeException e) {
                throw new ControllerException("None of the accepted media types is supported", HttpStatusCode.NotAcceptable);
            }

            // Obtain charset encoding
            String charset;
            try {
                charset = charsetResolver.getAcceptedCharset();
            } catch (UnsupportedEncodingException e) {
                throw new ControllerException("None of the accepted charset encodings is supported", e, HttpStatusCode.NotAcceptable);
            }

            // Send response headers
            response.sendHeaders(state.getStatus(), getHeadersWithContent(mediaType, charset));

            // Produce response body
            OutputStreamWriter writer;
            try {
                writer = new OutputStreamWriter(response.getBodyOutputStream(), charset);
            } catch (UnsupportedEncodingException e) {
                throw new ControllerException("None of the accepted charset encodings is supported", e, HttpStatusCode.NotAcceptable);
            }
            producer.produce(state.getResult(), writer);
        }
    }

    protected Map<String, String[]> getHeadersWithoutContent() throws ControllerException {
        Map<String, String[]> headers = new HashMap<String, String[]>();
        headers.put(HttpHeaders.CONTENT_LENGTH, new String[] { "0" } );
        return headers;
    }

    protected Map<String, String[]> getHeadersWithContent(String mediaType, String charset) throws ControllerException {
        Map<String, String[]> headers = new HashMap<String, String[]>();
        headers.put(HttpHeaders.CONTENT_LENGTH, new String[] { "-1" } );
        headers.put(HttpHeaders.CONTENT_TYPE, new String[] { mediaType + "; charset=" + charset } );
        return headers;
    }
}
