/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.controller.resolvers;

import net.craftforge.essential.core.constants.Charsets;
import net.craftforge.essential.core.constants.HttpHeaders;
import net.craftforge.essential.core.resolvers.HeaderResolver;
import net.craftforge.essential.core.utils.HeaderUtils;

import javax.inject.Inject;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;

/**
 * {@inheritDoc}
 *
 * @author Christian Bick
 * @since 28.11.11
 */
public class DefaultCharsetResolver implements CharsetResolver {

    private HeaderResolver headerResolver;

    @Inject
    public DefaultCharsetResolver(HeaderResolver headerResolver) {
        this.headerResolver = headerResolver;
    }

    /**
     * {@inheritDoc}
     *
     * Uses the charset given in the Content-Type header.
     *
     * Assumes UTF-8 as default if the header is not present or the charset is not specified.
     *
     * @return The content charset
     */
    public String getContentCharset() {
        return headerResolver.getHeaderValueAttribute(HttpHeaders.CONTENT_TYPE, "charset", Charsets.UTF8);
    }

    /**
     * {@inheritDoc}
     *
     * Uses the first supported charset given in the Accept-Charset header.
     *
     * Assumes UTF-8 as default if the header is not present.
     *
     * @return The accepted charset
     * @throws UnsupportedEncodingException if none of the accepted charset encodings is supported
     */
    public String getAcceptedCharset() throws UnsupportedEncodingException {
        String[] acceptCharsetHeaderValues = headerResolver.getHeaderValues(
                HttpHeaders.ACCEPT_CHARSET, new String[] { Charsets.UTF8 } );

        String[] acceptedCharsetEncodings = HeaderUtils.eliminateHeaderValuesAttributes(acceptCharsetHeaderValues);
        for (String acceptedCharset : acceptedCharsetEncodings) {
            acceptedCharset = acceptedCharset.trim();
            if (Charset.isSupported(acceptedCharset)) {
                return acceptedCharset;
            }
        }

        throw new UnsupportedEncodingException("None of the accepted charset encodings are supported: "
                + HeaderUtils.concatenateHeaderValues(acceptedCharsetEncodings));
    }
}
