/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.controller.resolvers;

import net.craftforge.essential.context.Request;
import net.craftforge.essential.core.constants.HttpHeaders;
import net.craftforge.essential.core.constants.MediaTypes;
import net.craftforge.essential.core.resolvers.HeaderResolver;
import net.craftforge.essential.core.utils.HeaderUtils;

import javax.inject.Inject;

/**
 * {@inheritDoc}
 *
 * @author Christian Bick
 * @since 28.11.11
 */
public class DefaultMediaTypeResolver implements MediaTypeResolver {

    private Request request;
    private HeaderResolver headerResolver;

    @Inject
    public DefaultMediaTypeResolver(Request request, HeaderResolver headerResolver) {
        this.request = request;
        this.headerResolver = headerResolver;
    }

    /**
     * {@inheritDoc}
     *
     * Uses the media type given in the Content-Type header.
     *
     * Assumes UTF-8 as default if the header is not present.
     *
     * @return The content media type
     */
    public String getContentMediaType() {
        String contentTypeHeader = headerResolver.getHeaderValue(HttpHeaders.CONTENT_TYPE, MediaTypes.TEXT_PLAIN);
        return HeaderUtils.eliminateHeaderValueAttributes(contentTypeHeader);
    }

    /**
     * {@inheritDoc}
     *
     * Uses the media types given in the the Accept header if not overridden in the URL meta part of
     * the URL (e.g. my/path.json).
     *
     * Assumes UTF-8 as default if the header is not present.
     *
     * @return The accepted media types
     */
    public String[] getAcceptedMediaTypes() {
        String metaDataPart = request.getUrlResourcePart();
        if (metaDataPart.contains(".text")) {
            return new String[] { MediaTypes.TEXT_PLAIN };
        } else if (metaDataPart.contains(".xml")) {
            return new String[] { MediaTypes.TEXT_XML };
        } else if (metaDataPart.contains(".json")) {
            return new String[] { MediaTypes.TEXT_JSON };
        } else {
            String[] acceptHeaderValues = headerResolver.getHeaderValues(
                    HttpHeaders.ACCEPT, new String[] { MediaTypes.TEXT_PLAIN });

            return HeaderUtils.eliminateHeaderValuesAttributes(acceptHeaderValues);
        }
    }
}
