/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.controller.resolvers;

import net.craftforge.essential.supply.Consumer;
import net.craftforge.essential.supply.exceptions.BadInputException;
import net.craftforge.essential.supply.exceptions.UnsupportedMediaTypeException;

import java.io.UnsupportedEncodingException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.List;

/**
 * An input resolver extracts the values for fields and
 * method parameters from an HTTP request.
 *
 * @author Christian Bick
 * @since 28.11.11
 */
public interface InputResolver {

    /**
     * Gets the input for a field from the corresponding request data (e.g. parameter, request body).
     * Uses the given consumer for request body deserialization.
     *
     * @param field The field to get the input for
     * @param consumer The consumer
     * @return The input for the field
     * @throws BadInputException if the request data for the input is not valid
     * @throws UnsupportedEncodingException if the request data's charset encoding is not supported
     * @throws UnsupportedMediaTypeException if the request data's media type is not supported
     */
    Object getInputForField(Field field, Consumer consumer)
            throws BadInputException, UnsupportedMediaTypeException, UnsupportedEncodingException;

    /**
     * Gets the input for a method's parameters from the corresponding request data (e.g. parameter, request body).
     * Uses the given consumer for request body deserialization.
     *
     * @param method The method to get the inputs for its parameters
     * @param consumer The consumer
     * @return The list of inputs (ordered by parameter order)
     * @throws BadInputException if the request data for the input is not valid
     * @throws UnsupportedEncodingException if the request data's charset encoding is not supported
     * @throws UnsupportedMediaTypeException if the request data's media type is not supported
     */
    List<Object> getInputForMethodParameters(Method method, Consumer consumer)
            throws BadInputException, UnsupportedMediaTypeException, UnsupportedEncodingException;
}
