/*
 * This file is part of essential.
 *
 *     essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with essential.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.craftforge.essential.controller.utils;

import net.craftforge.essential.controller.annotations.*;
import net.craftforge.essential.core.constants.HttpMethods;
import net.craftforge.essential.core.utils.UriUtils;
import net.craftforge.reflection.managers.ClassManager;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

/**
 * Utilities for annotation actions.
 *
 * @author Christian Bick
 * @since 17.02.2011
 */
public class ControllerReflUtils {

    /**
     * Gets the @Path annotation value of a class, looking it up on type level. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The @Path annotation value, null if not found
     */
    public static String getPathFromClass(Class<?> clazz) {
        Path annotation = (Path) ClassManager.getInstance(clazz).getTypeLevelAnnotation(Path.class);
        return annotation == null ? null : UriUtils.standardPath(annotation.value());
    }

    /**
     * Gets the @Path annotation value of a method, looking it up on method level. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param method The method
     * @return The @Path annotation value, null if not found
     */
    public static String getPathFromMethod(Method method) {
        Path annotation = (Path) ClassManager.getInstance(method).getMethodLevelAnnotation(method, Path.class);
        return annotation == null ? null : UriUtils.standardPath(annotation.value());
    }

    /**
     * Gets the @Param annotations value of a property, looking it up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param field The field
     * @return The @Param annotation value, null if not found
     */
    public static String getParamFromProperty(Field field) {
        Param annotation = (Param) ClassManager.getInstance(field).getPropertyLevelAnnotation(field, Param.class);
        return annotation == null ? null : annotation.value();
    }

    /**
     * Gets the @Header annotations value of a property, looking it up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param field The field
     * @return The @Header annotation value, null if not found
     */
    public static String getHeaderFromProperty(Field field) {
        Header annotation = (Header) ClassManager.getInstance(field).getPropertyLevelAnnotation(field, Header.class);
        return annotation == null ? null : annotation.value();
    }

    /**
     * Gets the @Property annotation value of a property, looking it up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param field The field
     * @return The @Property annotation value, null if not found
     */
    public static String getPropertyFromProperty(Field field) {
        Property annotation = (Property) ClassManager.getInstance(field).getPropertyLevelAnnotation(field, Property.class);
        return annotation == null ? null : annotation.value();
    }

    /**
     * Determines whether the @Body annotation is present on a property, looking it up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param field The field
     * @return The @Header annotation value, null if not found
     */
    public static boolean isBodyOnProperty(Field field) {
        Body annotation = (Body) ClassManager.getInstance(field).getPropertyLevelAnnotation(field, Body.class);
        return annotation != null;
    }

    /**
     * Gets the @DefaultValue annotation value(s) of a field, looking it up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param field The field
     * @return The @DefaultValue annotation value, null if not found
     */
    public static String[] getDefaultValuesFromProperty(Field field) {
        DefaultValue annotation =(DefaultValue) ClassManager.getInstance(field).getPropertyLevelAnnotation(field, DefaultValue.class);
        return annotation == null ? null : annotation.value();
    }

    /**
     * Gets the @Producer annotation value of a method, looking it up on method level first, then class level.
     * Uses the complete class hierarchy, including interfaces.
     *
     * @param method The method
     * @return The @Producer annotation value or null if not found
     */
    public static List<Class<?>> getProducersFromMethodOrClass(Method method) {
        Producers annotation = (Producers) ClassManager.getInstance(method).getMethodOrTypeLevelAnnotation(method, Producers.class);
        return annotation == null ? null : Arrays.asList(annotation.value());
    }

    /**
     * Gets the @Consumer annotation value of a method, looking it up on method level first, then class level.
     * Uses the complete class hierarchy, including interfaces.
     *
     * @param method The method
     * @return The @Consumer annotation value or null if not found
     */
    public static List<Class<?>> getConsumersFromMethodOrClass(Method method) {
        Consumers annotation = (Consumers) ClassManager.getInstance(method).getMethodOrTypeLevelAnnotation(method, Consumers.class);
        return annotation == null ? null : Arrays.asList(annotation.value());
    }

    /**
     * Checks if the @Public annotation is present, looking it up on method level first, then class level.
     * Uses the complete class hierarchy, including interfaces.
     *
     * @param method The method
     * @return Whether the @Public annotation is present or not
     */
    public static boolean isPublic(Method method) {
        Public annotation = (Public) ClassManager.getInstance(method).getMethodOrTypeLevelAnnotation(method, Public.class);
        return annotation != null;
    }

    /**
     * Gets a list of all fields with an @Param annotation, looking them up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The list of fields
     */
    public static List<Field> getParamFieldsFromClass(Class<?> clazz) {
        return ClassManager.getInstance(clazz).getAllPropertiesAnnotatedWith(Param.class);
    }

    /**
     * Gets a list of all fields with an @Header annotation, looking them up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The list of fields
     */
    public static List<Field> getHeaderFieldsFromClass(Class<?> clazz) {
        return ClassManager.getInstance(clazz).getAllPropertiesAnnotatedWith(Header.class);
    }

    /**
     * Gets a list of all fields with an @Body annotation, looking them up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The list of fields
     */
    public static List<Field> getBodyFieldsFromClass(Class<?> clazz) {
        return ClassManager.getInstance(clazz).getAllPropertiesAnnotatedWith(Body.class);
    }

    /**
     * Gets a list of all fields with an @Property annotation, looking them up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The list of fields
     */
    public static List<Field> getPropertyFieldsFromClass(Class<?> clazz) {
        return ClassManager.getInstance(clazz).getAllPropertiesAnnotatedWith(Property.class);
    }

    /**
     * Gets a list of all fields with an @Param, @Body, @Header or @Property annotation, looking them
     * up on field level and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The list of fields
     */
    public static List<Field> getAnnotatedFieldsFromClass(Class<?> clazz) {
        List<Field> annotatedFields = new LinkedList<Field>();
        annotatedFields.addAll(getParamFieldsFromClass(clazz));
        annotatedFields.addAll(getBodyFieldsFromClass(clazz));
        annotatedFields.addAll(getHeaderFieldsFromClass(clazz));
        annotatedFields.addAll(getPropertyFieldsFromClass(clazz));
        return annotatedFields;
    }

    /**
     * Gets a list of all methods with a REST annotation (@Get, @Post, etc.)
     * in the class, looking it up on method level. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The list of resource methods
     */
    public static List<Method> getResourceMethodsFromClass(Class<?> clazz) {
        List<Method> methods = new ArrayList<Method>();
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Get.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Post.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Put.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Delete.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Head.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Trace.class));
        return methods;
    }

    /**
     * Gets the REST annotation (@Get, @Post, etc.) of a method, looking it up on method
     * level. Uses the complete class hierarchy, including interfaces.
     *
     * @param method The method
     * @return The REST annotation or null if none found
     */
    public static String getHttpMethod(Method method) {
        Annotation annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Get.class);
        if (annotation != null) {
            return HttpMethods.GET;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Post.class);
        if (annotation != null) {
            return HttpMethods.POST;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Put.class);
        if (annotation != null) {
            return HttpMethods.PUT;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Delete.class);
        if (annotation != null) {
            return HttpMethods.DELETE;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Head.class);
        if (annotation != null) {
            return HttpMethods.HEAD;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Trace.class);
        if (annotation != null) {
            return HttpMethods.TRACE;
        }
        return null;
    }
}
